<?php

namespace jrosset\Collections;

use Closure;

/**
 * A collection
 *
 * @template TKey of array-key
 * @template TValue
 *
 * @implements ImmutableCollection<TKey, TValue>
 * @implements ICollection<TKey, TValue>
 */
class Collection extends ImmutableCollection implements ICollection {
    /**
     * @inheritDoc
     */
    public function set ($key, $value): static {
        return $this->_set($key, $value);
    }
    /**
     * @inheritDoc
     */
    public function merge (IImmutableCollection ...$collections): static {
        return $this->_merge(...$collections);
    }

    /**
     * @inheritDoc
     */
    public function add (...$values): static {
        return $this->_add(...$values);
    }
    /**
     * @inheritDoc
     */
    public function addCollection (IImmutableCollection ...$collections): static {
        return $this->_addCollection(...$collections);
    }

    /**
     * @inheritDoc
     */
    public function prepend (...$values): static {
        array_unshift($this->elements, ...$values);
        return $this;
    }
    /**
     * @inheritDoc
     */
    public function prependCollection (IImmutableCollection ...$collections): static {
        $prepend = array_merge(
            ...array_map(
                   function (IImmutableCollection $collection): array {
                       return $collection->values()->toArray();
                   },
                   $collections
               )
        );
        return $this->prepend(...$prepend);
    }

    /**
     * @inheritDoc
     */
    public function clear (): static {
        $this->_initialize();
        return $this;
    }
    /**
     * @inheritDoc
     */
    public function remove ($key): static {
        unset($this->elements[$this->_normalizeKey($key)]);
        return $this;
    }
    /**
     * @inheritDoc
     */
    public function removeValue ($value, bool $strict = false): static {
        foreach ($this->elements as $currentKey => $currentValue) {
            if (($strict && $value === $currentValue) || (!$strict && $value == $currentValue)) {
                unset($this->elements[$currentKey]);
            }
        }
        return $this;
    }

    /**
     * @inheritDoc
     */
    public function removeFirst () {
        $firstKey = $this->firstKey();
        $firstValue = $this->get($firstKey);

        $this->remove($firstKey);
        return $firstValue;
    }
    /**
     * @inheritDoc
     */
    public function removeLast () {
        $lastKey = $this->lastKey();
        $lastValue = $this->get($lastKey);

        $this->remove($lastKey);
        return $lastValue;
    }

    /**
     * @inheritDoc
     */
    public function offsetSet (mixed $offset, mixed $value): void {
        $this->set($offset, $value);
    }
    /**
     * @inheritDoc
     */
    public function offsetUnset (mixed $offset): void {
        $this->remove($offset);
    }

    /**
     * @inheritDoc
     */
    public function sliceSelf (int $offset, ?int $length = null): static {
        if ($this->count() === 0) {
            $this->elements = [];
        }
        else {
            $this->_checkOffset($offset);
            $this->elements = array_slice($this->elements, $offset, $length, true);
        }
        return $this;
    }

    /**
     * @inheritDoc
     */
    public function filterSelf (Closure $filter): static {
        $this->elements = array_filter($this->elements, $filter);
        return $this;
    }
    /**
     * @inheritDoc
     */
    public function removeEmpties (): static {
        return $this->filterSelf(function ($value) {
            return empty($value);
        });
    }
    /**
     * @inheritDoc
     */
    public function mapSelf (Closure $process): static {
        $this->elements = $this->map($process)->elements;
        return $this;
    }

    /**
     * @inheritDoc
     */
    public function sortSelf (Closure|IComparator|null $sorter = null): static {
        uasort($this->elements, self::_normalizeSorter($sorter));
        return $this;
    }
    /**
     * @inheritDoc
     */
    public function sortSelfByKey (Closure|IComparator $sorter): static {
        uksort($this->elements, self::_normalizeSorter($sorter));
        return $this;
    }
}