<?php

namespace jrosset\Collections;

use Closure;
use OutOfBoundsException;
use Throwable;

/**
 * Interface for a collection
 *
 * @template TKey of array-key
 * @template TValue
 *
 * @implements IImmutableCollection<TKey, TValue>
 */
interface ICollection extends IImmutableCollection {
    /**
     * Set a value to a key
     *
     * @param TKey   $key   The key
     * @param TValue $value The value
     *
     * @return $this
     */
    public function set ($key, $value): static;
    /**
     * Merge one or multiple collections into the current one
     *
     * @param IImmutableCollection<TKey, TValue> ...$collections The collections to merge
     *
     * @return $this
     */
    public function merge (IImmutableCollection ...$collections): static;

    /**
     * Add one or multiple values
     *
     * @param TValue ...$values The values
     *
     * @return $this
     */
    public function add (...$values): static;
    /**
     * Add the <b>values</b> of one or multiple collections
     *
     * @param IImmutableCollection<TKey, TValue> ...$collections The collections to add
     *
     * @return $this
     */
    public function addCollection (IImmutableCollection ...$collections): static;

    /**
     * Add one or multiple values at the beginning of the current collection
     *
     * @param TValue ...$values The values
     *
     * @return $this
     */
    public function prepend (...$values): static;
    /**
     * Add the <b>values</b> of one or multiple collections at the beginning of the current collection
     *
     * @param IImmutableCollection<TKey, TValue> ...$collections The collections to add
     *
     * @return $this
     */
    public function prependCollection (IImmutableCollection ...$collections): static;

    /**
     * Empty the collection
     *
     * @return $this
     */
    public function clear (): static;
    /**
     * Remove a value from it's key
     *
     * @param TKey $key The key
     *
     * @return $this
     */
    public function remove ($key): static;
    /**
     * Delete all instances of a value
     *
     * @param TValue $value  The value
     * @param bool   $strict Strict comparison ?
     *
     * @return $this
     */
    public function removeValue ($value, bool $strict = false): static;

    /**
     * Get and remove the first element
     *
     * @return TValue The value
     *
     * @throws OutOfBoundsException If the collection is empty
     */
    public function removeFirst ();
    /**
     * Get and remove the last element
     *
     * @return TValue The value
     *
     * @throws OutOfBoundsException If the collection is empty
     */
    public function removeLast ();

    /**
     * Keep only a slice of the collection
     *
     * Preserves the keys
     *
     * @param int      $offset The start offset
     * @param int|null $length The maximum length. Null if until end of the collection
     *
     * @return $this
     *
     * @throws OutOfBoundsException If the offset is not valid
     */
    public function sliceSelf (int $offset, ?int $length = null): static;

    /**
     * Keep only elements that satisfy predicate $filter
     *
     * @param Closure(TKey, TValue):bool $filter The filtering predicate
     *
     * @return $this
     */
    public function filterSelf (Closure $filter): static;
    /**
     * Keep only non-empty elements
     *
     * Use {@see https://www.php.net/manual/function.empty.php empty} to check if element is empty or not
     *
     * @return $this
     */
    public function removeEmpties (): static;
    /**
     * Applied <b>$process</b> on all elements
     *
     * @param Closure(TKey, TValue): TValue $process The process function to apply on each element
     *
     * @return $this
     */
    public function mapSelf (Closure $process): static;

    /**
     * Sort the elements (by value)
     *
     * @param null|IComparator|Closure(TValue, TValue):int $sorter The sorting method ; Null if values are object implementing {@see IComparable}.
     *                                                             <br>Return :
     *                                                             <br>&lt; 0 if value1 is before value2
     *                                                             <br>= 0 if value1 equals value2
     *                                                             <br>&gt; 0 if value1 is after value2
     *
     * @return $this
     *
     * @throws Throwable If an error occurs
     */
    public function sortSelf (Closure|IComparator|null $sorter = null): static;
    /**
     * Sort the elements by key
     *
     * @param IComparator|Closure(TKey, TKey):int $sorter The sorting method. Return :
     *                                                    <br>&lt; 0 if key1 is before key2
     *                                                    <br>= 0 if key1 equals key2
     *                                                    <br>&gt; 0 if key1 is after key2
     *
     * @return $this
     *
     * @throws Throwable If an error occurs
     */
    public function sortSelfByKey (Closure|IComparator $sorter): static;
}